// package myBinaryTree;

public class ArrayBinaryTree<E> {
	// ---------------------------------- data field
	E[] array;

	// ---------------------------------- method field

	// --------------------- constructors
	@SuppressWarnings("unchecked")
	public ArrayBinaryTree() {
		this.array = (E[]) new Object[10];
	}
	public ArrayBinaryTree(E root) {
		this.array = (E[]) new Object[10];
		this.array[0] = root;
	}
	public ArrayBinaryTree(E[] array) {
		this.array = array;
	}

	public void expandArray() {
		@SuppressWarnings("unchecked")
		E[] newArray = (E[]) new Object[this.array.length * 2];
		for(int i=0; i < this.array.length; i++) {
			newArray[i] = array[i];
		}
		this.array = newArray;
	}


	// --------------------- methods for element/node
	public E getElement(int nodeRank) {
		if(nodeRank < 0 || nodeRank >= this.array.length) {
			// System.out.println("Rank error!!!");
			return null;
		} else {
			return this.array[nodeRank];
		}
	}

	// left child
	public int getLeftChildRank(int nodeRank) {
		return 2 * nodeRank + 1;
	}
	public E getLeftChildElement(int nodeRank) {
		return this.getElement(2 * nodeRank + 1);
	}
	public void setLeftChildElement(int nodeRank, E element) {
		if(this.getLeftChildRank(nodeRank) >= this.array.length) {
			this.expandArray();	
		}
		this.array[this.getLeftChildRank(nodeRank)] = element;
	}
	// right child
	public int getRightChildRank(int nodeRank) {
		return 2 * nodeRank + 2;
	}
	public E getRightChildElement(int nodeRank) {
		return this.getElement(2 * nodeRank + 2);
	}
	public void setRightChildElement(int nodeRank, E element) {
		if(this.getRightChildRank(nodeRank) >= this.array.length) {
			this.expandArray();	
		}
		this.array[this.getRightChildRank(nodeRank)] = element;
	}
	
	// parent
	public boolean isLeftChild(int nodeRank) {
		return (nodeRank % 2 != 0);
	}
	public boolean isRightChild(int nodeRank) {
		return (nodeRank % 2 == 0);
	}

	public int getParentRank(int nodeRank) {
		if(this.isLeftChild(nodeRank)) {
			return (nodeRank-1)/2;
		} else {
			return (nodeRank-2)/2;
		}
	}
	public int getSiblingRank(int nodeRank) {
		if(this.isLeftChild(nodeRank)) {
			return nodeRank + 1;
		} else {
			return nodeRank - 1;
		}
	}

	// Tree methods
	public E getRoot(){
		return this.array[0];
	}
	public int size(int nodeRank) {
		if(this.getElement(nodeRank) == null) {
			return 0;
		}
		return 1 + this.size(this.getLeftChildRank(nodeRank)) + this.size(this.getRightChildRank(nodeRank));
	}

	public boolean isEmpty() {
		return (this.size(0) == 0);
	}

	public boolean isInternal(int nodeRank) {
		return (this.getLeftChildElement(nodeRank) != null || this.getRightChildElement(nodeRank) != null);
	}

	public boolean isExternal(int nodeRank) {
		return (this.getLeftChildElement(nodeRank) == null && this.getRightChildElement(nodeRank) == null);
	}

	public int height(int nodeRank) {
		if(this.isExternal(nodeRank)) {
			return 0;
		}
		else {
			return 1 + Math.max(height(this.getLeftChildRank(nodeRank)), height(this.getRightChildRank(nodeRank)));
		}
	}

	//Traversals
	public void inOrder(int nodeRank) {
		if(this.getLeftChildElement(nodeRank) != null) {
			inOrder(this.getLeftChildRank(nodeRank));
		}
		System.out.print(this.getElement(nodeRank)+ "	");
		if(this.getRightChildElement(nodeRank) != null) {
			inOrder(this.getRightChildRank(nodeRank));
		}
	}

	public void preOrder(ArrayBinaryTree<Character> tree, int nodeRank) {
		System.out.print(tree.getElement(nodeRank)+"	");
		if(tree.getLeftChildElement(nodeRank) != null) {
			preOrder(tree, tree.getLeftChildRank(nodeRank));
		}
		if(tree.getRightChildElement(nodeRank) != null) {
			preOrder(tree, tree.getRightChildRank(nodeRank));
		}
	}
	public void postOrder(ArrayBinaryTree<Character> tree, int nodeRank) {
		if(tree.getLeftChildElement(nodeRank) != null) {
			postOrder(tree, tree.getLeftChildRank(nodeRank));
		}
		if(tree.getRightChildElement(nodeRank) != null) {
			postOrder(tree, tree.getRightChildRank(nodeRank));
		}
		System.out.print(tree.getElement(nodeRank)+"	");
	}	

	public void display() {
		for(int i = 0; i < this.array.length; i++) {
			System.out.print(this.array[i] + " ");
		}
		System.out.println();
	}

	// public static void main(String[] args) {
	// 	// ------------ test ArrayBinaryTree
	// 	System.out.println("============================= Test ArrayBinaryTree");
	// 	Character[] array = {'+', 'x', 'x', '2', '-', '3', 'b', null, null,  'a', '1'};
	// 	ArrayBinaryTree<Character> arrayTree = new ArrayBinaryTree<Character>(array);
		
	// 	System.out.println("Size of the tree: " + arrayTree.size(0));
	// 	System.out.println("Height of the tree: " + arrayTree.height(0));
	// 	System.out.println("Height of node1: " + arrayTree.height(10));
	// 	System.out.println("Height of nodeminus: " + arrayTree.height(4));

	// 	System.out.print("\npreOrder : ");
	// 	arrayTree.preOrder(arrayTree, 0);
	// 	System.out.print("\ninOrder  : ");
		// arrayTree.inOrder(0);
	// 	System.out.print("\npostOrder: ");
	// 	arrayTree.postOrder(arrayTree, 0);

	// 	System.out.println("\n\n");
	// }
}
